@@ -8,18 +8,23 @@
 package XML::SAX::Writer;
 use strict;
 use Text::Iconv             qw();
-use XML::NamespaceSupport   qw();
 use XML::SAX::Exception     qw();
+use XML::SAX::Writer::XML   qw();
+use XML::Filter::BufferText qw();
 @XML::SAX::Writer::Exception::ISA = qw(XML::SAX::Exception);
 
-use vars qw($VERSION %DEFAULT_ESCAPE);
-$VERSION = '0.38';
+use vars qw($VERSION %DEFAULT_ESCAPE %COMMENT_ESCAPE);
+$VERSION = '0.41';
+
 %DEFAULT_ESCAPE = (
                     '&'     => '&amp;',
                     '<'     => '&lt;',
                     '>'     => '&gt;',
                     '"'     => '&quot;',
                     "'"     => '&apos;',
+                  );
+
+%COMMENT_ESCAPE = (
                     '--'    => '&#45;&#45;',
                   );
 
@@ -32,55 +37,67 @@ sub new {
     my $opt   = (@_ == 1)  ? { %{shift()} } : {@_};
 
     # default the options
-    $opt->{Escape}      ||= \%DEFAULT_ESCAPE;
-    $opt->{EncodeFrom}  ||= 'utf-8';
-    $opt->{EncodeTo}    ||= 'utf-8';
-    $opt->{Format}      ||= {}; # needs options w/ defaults, we'll see later
-    $opt->{Output}      ||= *{STDOUT}{IO};
+    $opt->{Writer}          ||= 'XML::SAX::Writer::XML';
+    $opt->{Escape}          ||= \%DEFAULT_ESCAPE;
+    $opt->{CommentEscape}   ||= \%COMMENT_ESCAPE;
+    $opt->{EncodeFrom}      ||= 'utf-8';
+    $opt->{EncodeTo}        ||= 'utf-8';
+    $opt->{Format}          ||= {}; # needs options w/ defaults, we'll see later
+    $opt->{Output}          ||= *{STDOUT}{IO};
+    eval "use $opt->{Writer};";
 
-    return bless $opt, $class;
+    my $obj = bless $opt, $opt->{Writer};
+    $obj->init;
+
+    # we need to buffer the text to escape it right
+    my $bf = XML::Filter::BufferText->new( Handler => $obj );
+
+    return $bf;
 }
 #-------------------------------------------------------------------#
 
-
-#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,#
-#`,`, The SAX Handler `,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,#
-#```````````````````````````````````````````````````````````````````#
+#-------------------------------------------------------------------#
+# init
+#-------------------------------------------------------------------#
+sub init {} # noop, for subclasses
+#-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
-# start_document
+# setConverter
 #-------------------------------------------------------------------#
-sub start_document {
+sub setConverter {
     my $self = shift;
 
-    # init the object
     if (lc($self->{EncodeFrom}) ne lc($self->{EncodeTo})) {
         $self->{Encoder} = Text::Iconv->new($self->{EncodeFrom}, $self->{EncodeTo});
     }
     else {
         $self->{Encoder} = XML::SAX::Writer::NullConverter->new;
     }
+    return $self;
+}
+#-------------------------------------------------------------------#
 
-    $self->{EscaperRegex} = eval 'qr/'                                                .
-                            join( '|', map { $_ = "\Q$_\E" } keys %{$self->{Escape}}) .
-                            '/;'                                                  ;
-
-    $self->{NSDecl} = [];
-    $self->{NSHelper} = XML::NamespaceSupport->new({ xmlns => 1, fatal_errors => 0 });
-    $self->{NSHelper}->pushContext;
-
+#-------------------------------------------------------------------#
+# setConsumer
+#-------------------------------------------------------------------#
+sub setConsumer {
+    my $self = shift;
 
     # create the Consumer
     my $ref = ref $self->{Output};
     if ($ref eq 'SCALAR') {
         $self->{Consumer} = XML::SAX::Writer::StringConsumer->new($self->{Output});
     }
+    elsif ($ref eq 'CODE') {
+        $self->{Consumer} = XML::SAX::Writer::CodeConsumer->new($self->{Output});
+    }
     elsif ($ref eq 'ARRAY') {
         $self->{Consumer} = XML::SAX::Writer::ArrayConsumer->new($self->{Output});
     }
     elsif (
-            $ref eq 'GLOB'                                or
-            UNIVERSAL::isa($self->{Output}, 'GLOB')       or
+#            $ref eq 'GLOB'                                or
+            UNIVERSAL::isa(\$self->{Output}, 'GLOB')      or
             UNIVERSAL::isa($self->{Output}, 'IO::Handle')) {
         $self->{Consumer} = XML::SAX::Writer::HandleConsumer->new($self->{Output});
     }
@@ -93,495 +110,41 @@ sub start_document {
     else {
         XML::SAX::Writer::Exception->throw( Message => 'Unknown option for Output' );
     }
+    return $self;
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
-# end_document
+# setEscaperRegex
 #-------------------------------------------------------------------#
-sub end_document {
-    my $self = shift;
-    # we may need to do a little more here
-    $self->{NSHelper}->popContext;
-    return $self->{Consumer}->finalize;
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# start_element
-#-------------------------------------------------------------------#
-sub start_element {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-    my $attr = $data->{Attributes};
-
-    # fix the namespaces and prefixes of what we're receiving, in case
-    # something is wrong
-    if ($data->{NamespaceURI}) {
-        my $uri = $self->{NSHelper}->getURI($data->{Prefix});
-        if ($uri ne $data->{NamespaceURI}) { # ns has precedence
-            $data->{Prefix} = $self->{NSHelper}->getPrefix($data->{NamespaceURI}); # random, but correct
-            $data->{Name} = $data->{Prefix} ? "$data->{Prefix}:$data->{LocalName}" : "$data->{LocalName}";
-        }
-    }
-    elsif ($data->{Prefix}) { # we can't have a prefix and no NS
-        $data->{Name}   = $data->{LocalName};
-        $data->{Prefix} = '';
-    }
-
-    # create a hash containing the attributes so that we can ensure there is
-    # no duplication. Also, we check that ns are properly declared, that the
-    # Name is good, etc...
-    my %attr_hash;
-    for my $at (values %$attr) {
-        if ($at->{NamespaceURI}) {
-            my $uri = $self->{NSHelper}->getURI($at->{Prefix});
-            if ($uri ne $at->{NamespaceURI}) { # ns has precedence
-                $at->{Prefix} = $self->{NSHelper}->getPrefix($at->{NamespaceURI}); # random, but correct
-                $at->{Name} = $at->{Prefix} ? "$at->{Prefix}:$at->{LocalName}" : "$at->{LocalName}";
-            }
-        }
-        elsif ($at->{Prefix}) { # we can't have a prefix and no NS
-            $at->{Name}   = $at->{LocalName};
-            $at->{Prefix} = '';
-        }
-        $attr_hash{$at->{Name}} = $at->{Value};
-    }
-
-    for my $nd (@{$self->{NSDecl}}) {
-        if ($nd->{Prefix}) {
-            $attr_hash{'xmlns:' . $nd->{Prefix}} = $nd->{NamespaceURI};
-        }
-        else {
-            $attr_hash{'xmlns'} = $nd->{NamespaceURI};
-        }
-    }
-    $self->{NSDecl} = [];
-
-    # build a string from what we have, and buffer it
-    my $el = '<' . $data->{Name};
-    for my $k (keys %attr_hash) {
-        $el .= ' ' . $k . '=\'' . $self->_escape($attr_hash{$k}) . '\'';
-    }
-
-    $self->{BufferElement} = $el;
-    $self->{NSHelper}->pushContext;
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# end_element
-#-------------------------------------------------------------------#
-sub end_element {
-    my $self = shift;
-    my $data = shift;
-
-    my $el;
-    if ($self->{BufferElement}) {
-        $el = $self->{BufferElement} . ' />';
-    }
-    else {
-        $el = '</' . $data->{Name} . '>';
-    }
-    $el = $self->{Encoder}->convert($el);
-    $self->{Consumer}->output($el);
-    $self->{NSHelper}->popContext;
-    $self->{BufferElement} = '';
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# characters
-#-------------------------------------------------------------------#
-sub characters {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-
-    my $char = $data->{Data};
-    $char = $self->_escape($char) unless $self->{InCDATA};
-    $char = $self->{Encoder}->convert($char);
-    $self->{Consumer}->output($char);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# start_prefix_mapping
-#-------------------------------------------------------------------#
-sub start_prefix_mapping {
-    my $self = shift;
-    my $data = shift;
-
-    push @{$self->{NSDecl}}, $data;
-    $self->{NSHelper}->declarePrefix($data->{Prefix}, $data->{NamespaceURI});
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# end_prefix_mapping
-#-------------------------------------------------------------------#
-sub end_prefix_mapping {}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# processing_instruction
-#-------------------------------------------------------------------#
-sub processing_instruction {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-    $self->_output_dtd;
-
-    my $pi = "<?$data->{Target} $data->{Data}?>";
-    $pi = $self->{Encoder}->convert($pi);
-    $self->{Consumer}->output($pi);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# ignorable_whitespace
-#-------------------------------------------------------------------#
-sub ignorable_whitespace {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-
-    my $char = $data->{Data};
-    $char = $self->_escape($char);
-    $char = $self->{Encoder}->convert($char);
-    $self->{Consumer}->output($char);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# skipped_entity
-#-------------------------------------------------------------------#
-sub skipped_entity {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-    $self->_output_dtd;
-
-    my $ent;
-    if ($data->{Name} =~ m/^%/) {
-        $ent = $data->{Name} . ';';
-    }
-    else {
-        $ent = '&' . $data->{Name} . ';';
-    }
-
-    $ent = $self->{Encoder}->convert($ent);
-    $self->{Consumer}->output($ent);
-
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# notation_decl
-#-------------------------------------------------------------------#
-sub notation_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $not = "    <!NOTATION " . $data->{Name};
-    if ($data->{PublicId} and $data->{SystemId}) {
-        $not .= ' PUBLIC \'' . $self->_escape($data->{PublicId}) . '\' \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    elsif ($data->{PublicId}) {
-        $not .= ' PUBLIC \'' . $self->_escape($data->{PublicId}) . '\'';
-    }
-    else {
-        $not .= ' SYSTEM \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    $not .= " >\n";
-
-    $not = $self->{Encoder}->convert($not);
-    $self->{Consumer}->output($not);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# unparsed_entity_decl
-#-------------------------------------------------------------------#
-sub unparsed_entity_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $ent = "    <!ENTITY " . $data->{Name};
-    if ($data->{PublicId}) {
-        $ent .= ' PUBLIC \'' . $self->_escape($data->{PublicId}) . '\' \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    else {
-        $ent .= ' SYSTEM \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    $ent .= " NDATA $data->{Notation} >\n";
-
-
-    $ent = $self->{Encoder}->convert($ent);
-    $self->{Consumer}->output($ent);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# element_decl
-#-------------------------------------------------------------------#
-sub element_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $eld = "    <!ELEMENT " . $data->{Name} . ' ' . $data->{Model} . " >\n";
-    $eld = $self->{Encoder}->convert($eld);
-    $self->{Consumer}->output($eld);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# attribute_decl
-#-------------------------------------------------------------------#
-sub attribute_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $atd = "      <!ATTLIST " . $data->{eName} . ' ' . $data->{aName} . ' ';
-    $atd   .= $data->{Type} . ' ' . $data->{ValueDefault} . ' ';
-    $atd   .= $data->{Value} . ' ' if $data->{Value};
-    $atd   .= " >\n";
-    $atd = $self->{Encoder}->convert($atd);
-    $self->{Consumer}->output($atd);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# internal_entity_decl
-#-------------------------------------------------------------------#
-sub internal_entity_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $ent = "    <!ENTITY " . $data->{Name} . ' \'' . $self->_escape($data->{Value}) . "' >\n";
-    $ent = $self->{Encoder}->convert($ent);
-    $self->{Consumer}->output($ent);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# external_entity_decl
-#-------------------------------------------------------------------#
-sub external_entity_decl {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_dtd;
-
-    # I think that param entities are normalized before this
-    my $ent = "    <!ENTITY " . $data->{Name};
-    if ($data->{PublicId}) {
-        $ent .= ' PUBLIC \'' . $self->_escape($data->{PublicId}) . '\' \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    else {
-        $ent .= ' SYSTEM \'' . $self->_escape($data->{SystemId}) . '\'';
-    }
-    $ent .= " >\n";
-
-
-    $ent = $self->{Encoder}->convert($ent);
-    $self->{Consumer}->output($ent);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# comment
-#-------------------------------------------------------------------#
-sub comment {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-    $self->_output_dtd;
-
-    my $cmt = '<!--' . $self->_escape($data->{Data}) . '-->';
-    $cmt = $self->{Encoder}->convert($cmt);
-    $self->{Consumer}->output($cmt);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# start_dtd
-#-------------------------------------------------------------------#
-sub start_dtd {
-    my $self = shift;
-    my $data = shift;
-
-    my $dtd = '<!DOCTYPE ' . $data->{Name};
-    if ($data->{PublicId}) {
-        $dtd .= ' PUBLIC \'' . $self->_escape($data->{PublicId}) . '\' \'' . $self->_escape($data->{SysmteId}) . '\'';
-    }
-    elsif ($data->{SystemId}) {
-        $dtd .= ' SYSTEM \'' . $self->_escape($data->{SysmteId}) . '\'';
-    }
-
-    $self->{BufferDTD} = $dtd;
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# end_dtd
-#-------------------------------------------------------------------#
-sub end_dtd {
-    my $self = shift;
-    my $data = shift;
-
-    my $dtd;
-    if ($self->{BufferDTD}) {
-        $dtd = $self->{BufferDTD} . ' >';
-    }
-    else {
-        $dtd = ' ]>';
-    }
-    $dtd = $self->{Encoder}->convert($dtd);
-    $self->{Consumer}->output($dtd);
-    $self->{BufferDTD} = '';
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# start_cdata
-#-------------------------------------------------------------------#
-sub start_cdata {
-    my $self = shift;
-    $self->_output_element;
-
-    $self->{InCDATA} = 1;
-    my $cds = $self->{Encoder}->convert('<![CDATA[');
-    $self->{Consumer}->output($cds);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# end_cdata
-#-------------------------------------------------------------------#
-sub end_cdata {
+sub setEscaperRegex {
     my $self = shift;
 
-    $self->{InCDATA} = 0;
-    my $cds = $self->{Encoder}->convert(']]>');
-    $self->{Consumer}->output($cds);
+    $self->{EscaperRegex} = eval 'qr/'                                                .
+                            join( '|', map { $_ = "\Q$_\E" } keys %{$self->{Escape}}) .
+                            '/;'                                                  ;
+    return $self;
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
-# start_entity
+# setCommentEscaperRegex
 #-------------------------------------------------------------------#
-sub start_entity {
-    my $self = shift;
-    my $data = shift;
-    $self->_output_element;
-    $self->_output_dtd;
-
-    my $ent;
-    if ($data->{Name} eq '[dtd]') {
-        # we ignore the fact that we're dealing with an external
-        # DTD entity here, and prolly shouldn't write the DTD
-        # events unless explicitly told to
-        # this will prolly change
-    }
-    elsif ($data->{Name} =~ m/^%/) {
-        $ent = $data->{Name} . ';';
-    }
-    else {
-        $ent = '&' . $data->{Name} . ';';
-    }
-
-    $ent = $self->{Encoder}->convert($ent);
-    $self->{Consumer}->output($ent);
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# end_entity
-#-------------------------------------------------------------------#
-sub end_entity {
-    # depending on what is done above, we might need to do sth here
-}
-#-------------------------------------------------------------------#
-
-
-### SAX1 stuff ######################################################
-
-#-------------------------------------------------------------------#
-# xml_decl
-#-------------------------------------------------------------------#
-sub xml_decl {
-    my $self = shift;
-    my $data = shift;
-
-    # version info is compulsory, contrary to what some seem to think
-    # also, there's order in the pseudo-attr
-    my $xd = '';
-    if ($data->{Version}) {
-        $xd .= "<?xml version='$data->{Version}'";
-        if ($data->{Encoding}) {
-            $xd .= " encoding='$data->{Encoding}'";
-        }
-        if ($data->{Standalone}) {
-            $xd .= " standalone='$data->{Standalone}'";
-        }
-        $xd .= '?>';
-    }
-
-    $xd = $self->{Encoder}->convert($xd); # this may blow up
-    $self->{Consumer}->output($xd);
-}
-#-------------------------------------------------------------------#
-
-
-#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,#
-#`,`, Helpers `,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,#
-#```````````````````````````````````````````````````````````````````#
-
-#-------------------------------------------------------------------#
-# _output_element
-#-------------------------------------------------------------------#
-sub _output_element {
+sub setCommentEscaperRegex {
     my $self = shift;
 
-    if ($self->{BufferElement}) {
-        my $el = $self->{BufferElement} . '>';
-        $el = $self->{Encoder}->convert($el);
-        $self->{Consumer}->output($el);
-        $self->{BufferElement} = '';
-    }
+    $self->{CommentEscaperRegex} =
+                        eval 'qr/'                                                .
+                        join( '|', map { $_ = "\Q$_\E" } keys %{$self->{CommentEscape}}) .
+                        '/;'                                                  ;
+    return $self;
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
-# _output_dtd
+# escape
 #-------------------------------------------------------------------#
-sub _output_dtd {
-    my $self = shift;
-
-    if ($self->{BufferDTD}) {
-        my $dtd = $self->{BufferDTD} . " [\n";
-        $dtd = $self->{Encoder}->convert($dtd);
-        $self->{Consumer}->output($dtd);
-        $self->{BufferDTD} = '';
-    }
-}
-#-------------------------------------------------------------------#
-
-#-------------------------------------------------------------------#
-# _escape
-#-------------------------------------------------------------------#
-sub _escape {
+sub escape {
     my $self = shift;
     my $str  = shift;
 
@@ -590,10 +153,17 @@ sub _escape {
 }
 #-------------------------------------------------------------------#
 
+#-------------------------------------------------------------------#
+# escapeComment
+#-------------------------------------------------------------------#
+sub escapeComment {
+    my $self = shift;
+    my $str  = shift;
 
-
-
-
+    $str =~ s/($self->{CommentEscaperRegex})/$self->{CommentEscape}->{$1}/oge;
+    return $str;
+}
+#-------------------------------------------------------------------#
 
 
 
@@ -605,7 +175,15 @@ sub _escape {
 # new methods are added to the interface
 
 package XML::SAX::Writer::ConsumerInterface;
-sub new {}
+
+sub new {
+    my $class = shift;
+    my $ref = shift;
+    ## $self is a reference to the reference that we will send output
+    ## to.  This allows us to bless $self without blessing $$self.
+    return bless \$ref, ref $class || $class;
+}
+
 sub output {}
 sub finalize {}
 
@@ -615,33 +193,57 @@ sub finalize {}
 #```````````````````````````````````````````````````````````````````#
 
 package XML::SAX::Writer::StringConsumer;
-use base qw(XML::SAX::Writer::ConsumerInterface);
+@XML::SAX::Writer::StringConsumer::ISA = qw(XML::SAX::Writer::ConsumerInterface);
 
 #-------------------------------------------------------------------#
 # new
 #-------------------------------------------------------------------#
 sub new {
-    my $class = ref($_[0]) ? ref(shift) : shift;
-    my $str   = shift;
-    $$str = '';
-    return bless [$str], $class;
+    my $self = shift->SUPER::new( @_ );
+    $$self = '';
+    return $self;
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
 # output
 #-------------------------------------------------------------------#
-sub output {
-    my $self = shift;
-    my $data = shift;
-    ${$self->[0]} .= $data;
-}
+sub output { ${$_[0]}.= pop }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
 # finalize
 #-------------------------------------------------------------------#
-sub finalize { return $_[0]->[0]; }
+sub finalize { $_[0] }
+#-------------------------------------------------------------------#
+
+#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,#
+#`,`, The Code Consumer `,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,`,#
+#```````````````````````````````````````````````````````````````````#
+
+package XML::SAX::Writer::CodeConsumer;
+@XML::SAX::Writer::CodeConsumer::ISA = qw(XML::SAX::Writer::ConsumerInterface );
+
+#-------------------------------------------------------------------#
+# new
+#-------------------------------------------------------------------#
+sub new {
+    my $self = shift->SUPER::new( @_ );
+    $$self->( 'start_document', '' );
+    return $self;
+}
+#-------------------------------------------------------------------#
+
+#-------------------------------------------------------------------#
+# output
+#-------------------------------------------------------------------#
+sub output { ${$_[0]}->('data', pop) } ## Avoid an extra copy
+#-------------------------------------------------------------------#
+
+#-------------------------------------------------------------------#
+# finalize
+#-------------------------------------------------------------------#
+sub finalize { ${$_[0]}->('end_document', '') }
 #-------------------------------------------------------------------#
 
 
@@ -650,33 +252,28 @@ sub finalize { return $_[0]->[0]; }
 #```````````````````````````````````````````````````````````````````#
 
 package XML::SAX::Writer::ArrayConsumer;
-use base qw(XML::SAX::Writer::ConsumerInterface);
+@XML::SAX::Writer::ArrayConsumer::ISA = qw(XML::SAX::Writer::ConsumerInterface);
 
 #-------------------------------------------------------------------#
 # new
 #-------------------------------------------------------------------#
 sub new {
-    my $class = ref($_[0]) ? ref(shift) : shift;
-    my $arr   = shift;
-    @$arr = ();
-    return bless [$arr], $class;
+    my $self = shift->SUPER::new( @_ );
+    @$$self = ();
+    return $self;
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
 # output
 #-------------------------------------------------------------------#
-sub output {
-    my $self = shift;
-    my $data = shift;
-    push @{$self->[0]}, $data;
-}
+sub output { push @${$_[0]}, pop }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
 # finalize
 #-------------------------------------------------------------------#
-sub finalize { return $_[0]->[0]; }
+sub finalize { return ${$_[0]} }
 #-------------------------------------------------------------------#
 
 
@@ -685,37 +282,23 @@ sub finalize { return $_[0]->[0]; }
 #```````````````````````````````````````````````````````````````````#
 
 package XML::SAX::Writer::HandleConsumer;
-use base qw(XML::SAX::Writer::ConsumerInterface);
-
-#-------------------------------------------------------------------#
-# new
-#-------------------------------------------------------------------#
-sub new {
-    my $class = ref($_[0]) ? ref(shift) : shift;
-    my $fh    = shift;
-    return bless [$fh], $class;
-}
-#-------------------------------------------------------------------#
+@XML::SAX::Writer::HandleConsumer::ISA = qw(XML::SAX::Writer::ConsumerInterface);
 
 #-------------------------------------------------------------------#
 # output
 #-------------------------------------------------------------------#
 sub output {
-    my $self = shift;
-    my $data = shift;
-    my $fh = $self->[0];
-    print $fh $data or XML::SAX::Exception->throw( Message => "Could not write to handle: $fh ($!)" );
+    my $fh = ${$_[0]};
+    print $fh pop or XML::SAX::Exception->throw(
+        Message => "Could not write to handle: $fh ($!)"
+    );
 }
 #-------------------------------------------------------------------#
 
 #-------------------------------------------------------------------#
 # finalize
 #-------------------------------------------------------------------#
-sub finalize {
-    my $self = shift;
-#    close $self;
-    return 0;
-}
+sub finalize { return 0 }
 #-------------------------------------------------------------------#
 
 
@@ -724,25 +307,25 @@ sub finalize {
 #```````````````````````````````````````````````````````````````````#
 
 package XML::SAX::Writer::FileConsumer;
-use base qw(XML::SAX::Writer::HandleConsumer);
-#use Symbol  qw();
+@XML::SAX::Writer::FileConsumer::ISA = qw(XML::SAX::Writer::HandleConsumer);
 
 #-------------------------------------------------------------------#
 # new
 #-------------------------------------------------------------------#
 sub new {
-    my $class = ref($_[0]) ? ref(shift) : shift;
-    my $file  = shift or XML::SAX::Writer::Exception->throw(
-                            Message => "No filename provided to XML::SAX::Writer::FileConsumer"
-                                                            );
-#    my $sym = Symbol::gensym;
+    my ( $proto, $file ) = ( shift, shift );
+
+    XML::SAX::Writer::Exception->throw(
+        Message => "No filename provided to " . ref( $proto || $proto )
+    ) unless defined $file;
+
     local *XFH;
+
     open XFH, ">$file" or XML::SAX::Writer::Exception->throw(
-                                    Message => "Error opening file $file: $!"
-                                                            );
-    my $self = XML::SAX::Writer::HandleConsumer->new(*{XFH}{IO});
-    return bless $self, $class;
-#    return SUPER->new(\*$sym);
+        Message => "Error opening file $file: $!"
+    );
+
+    return $proto->SUPER::new( *{XFH}{IO}, @_ );
 }
 #-------------------------------------------------------------------#
 
@@ -750,8 +333,7 @@ sub new {
 # finalize
 #-------------------------------------------------------------------#
 sub finalize {
-    my $self = shift;
-    close $self->[0];
+    close ${$_[0]};
     return 0;
 }
 #-------------------------------------------------------------------#
@@ -763,7 +345,7 @@ sub finalize {
 
 package XML::SAX::Writer::NullConverter;
 sub new     { return bless [], __PACKAGE__ }
-sub convert {}
+sub convert { $_[1] }
 
 
 1;
@@ -775,7 +357,7 @@ sub convert {}
 
 =head1 NAME
 
-XML::SAX::Writer - SAX2 XML Writer
+XML::SAX::Writer - SAX2 Writer
 
 =head1 SYNOPSIS
 
@@ -839,15 +421,20 @@ If this parameter is not provided, then output is sent to STDOUT.
 This should be a hash reference where the keys are characters
 sequences that should be escaped and the values are the escaped form
 of the sequence. By default, this module will escape the ampersand
-(&), less than (<), greater than (>), double quote ("), apostrophe
-('), and double dash (--) character sequences. Note that the double
-dash escape is needed for comments, and that some browsers don't
-support the &apos; escape used for apostrophes so that you should
-be careful when outputting XHTML.
+(&), less than (<), greater than (>), double quote ("), and apostrophe
+('). Note that some browsers don't support the &apos; escape used for
+apostrophes so that you should be careful when outputting XHTML.
 
 If you only want to add entries to the Escape hash, you can first
 copy the contents of %XML::SAX::Writer::DEFAULT_ESCAPE.
 
+=item -- CommentEscape
+
+Comment content often needs to be escaped differently from other
+content. This option works exactly as the previous one except that
+by default it only escapes the double dash (--) and that the contents
+can be copied from %XML::SAX::Writer::COMMENT_ESCAPE.
+
 =item -- EncodeFrom
 
 The character set encoding in which incoming data will be provided.
@@ -860,12 +447,10 @@ this defaults to UTF-8.
 
 =back
 
-=head1 GENERATING XML
-
 =head1 THE CONSUMER INTERFACE
 
 XML::SAX::Writer can receive pluggable consumer objects that will be
-in charge of writing out the XML formatted by this module. Setting a
+in charge of writing out what is formatted by this module. Setting a
 Consumer is done by setting the Output option to the object of your
 choice instead of to an array, scalar, or file handle as is more
 commonly done (internally those in fact map to Consumer classes and
@@ -884,7 +469,9 @@ The two methods that it needs to implement are:
 
 =over 4
 
-=item * output(String)
+=item * output STRING
+
+(Required)
 
 This is called whenever the Writer wants to output a string formatted
 in XML. Encoding conversion, character escaping, and formatting have
@@ -893,6 +480,8 @@ with the string.
 
 =item * finalize()
 
+(Optional)
+
 This is called once the document has been output in its entirety,
 during the end_document event. end_document will in fact return
 whatever finalize() returns, and that in turn should be returned
@@ -901,8 +490,129 @@ you need to provide feedback of some sort.
 
 =back
 
+Here's an example of a custom consumer.  Note the extra C<$> signs in
+front of $self; the base class is optimized for the overwhelmingly
+common case where only one data member is required and $self is a
+reference to that data member.
+
+    package MyConsumer;
+
+    @ISA = qw( XML::SAX::Writer::ConsumerInterface );
+
+    use strict;
+
+    sub new {
+        my $self = shift->SUPER::new( my $output );
+
+        $$self = '';      # Note the extra '$'
+
+        return $self;
+    }
+
+    sub output {
+        my $self = shift;
+        $$self .= uc shift;
+    }
+
+    sub get_output {
+        my $self = shift;
+        return $$self;
+    }
+
+And here's one way to use it:
+
+    my $c = MyConsumer->new;
+    my $w = XML::SAX::Writer->new( Output => $c );
+
+    ## ... send events to $w ...
+
+    print $c->get_output;
+
+If you need to store more that one data member, pass in an array or hash
+reference:
+
+        my $self = shift->SUPER::new( {} );
+
+and access it like:
+
+    sub output {
+        my $self = shift;
+        $$self->{Output} .= uc shift;
+    }
+
+=head1 THE ENCODER INTERFACE
+
+Encoders can be plugged in to allow one to use one's favourite encoder
+object. Presently there are two encoders: Iconv and NullEncoder, and
+one based on C<Encode> ought to be out soon. They need to implement
+two methods, and may inherit from XML::SAX::Writer::NullConverter if
+they wish to
+
+=over 4
+
+=item new FROM_ENCODING, TO_ENCODING
+
+Creates a new Encoder. The arguments are the chosen encodings.
+
+=item convert STRING
+
+Converts that string and returns it.
+
+=back
+
+=head1 CUSTOM OUTPUT
+
+This module is generally used to write XML -- which it does most of the
+time -- but just like the rest of SAX it can be used as a generic
+framework to output data, the opposite of a non-XML SAX parser.
+
+Of course there's only so much that one can abstract, so depending on
+your format this may or may not be useful. If it is, you'll need to
+know the followin API (and probably to have a look inside
+C<XML::SAX::Writer::XML>, the default Writer).
+
+=over
+
+=item init
+
+Called before the writing starts, it's a chance for the subclass to do
+some initialisation if it needs it.
+
+=item setConverter
+
+This is used to set the proper converter for character encodings. The
+default implementation should suffice but you can override it. It must
+set C<$self->{Encoder}> to an Encoder object. Subclasses *should* call
+it.
+
+=item setConsumer
+
+Same as above, except that it is for the Consumer object, and that it
+must set C<$self->{Consumer}>.
+
+=item setEscaperRegex
+
+Will initialise the escaping regex C<$self->{EscaperRegex}> based on
+what is needed.
+
+=item escape STRING
+
+Takes a string and escapes it properly.
+
+=item setCommentEscaperRegex and escapeComment STRING
+
+These work exactly the same as the two above, except that they are meant
+to operate on comment contents, which often have different escaping rules
+than those that apply to regular content.
+
+=back
+
 =head1 TODO
 
+    - proper UTF-16 handling
+
+    - deprecation of xml_decl and addition of the logic to replace it
+
     - make the quote character an option. By default it is here ', but
     I know that a lot of people (for reasons I don't understand but
     won't question :-) prefer to use ". (on most keyboards " is more
@@ -921,10 +631,14 @@ you need to provide feedback of some sort.
     and detecter both in Makefile.PL requirements and in the module
     at runtime.
 
+    - remove the xml_decl and replace it with intelligent logic, as
+    discussed on perl-xml
+
 =head1 CREDITS
 
-Michael Koehne (XML::Handler::YAWriter) for much inspiration and
-Barrie Slaymaker for the Consumer pattern idea. Of course the usual
+Michael Koehne (XML::Handler::YAWriter) for much inspiration and Barrie
+Slaymaker for the Consumer pattern idea, the coderef output option and
+miscellaneous bugfixes and performance tweaks. Of course the usual
 suspects (Kip Hampton and Matt Sergeant) helped in the usual ways.
 
 =head1 AUTHOR
